/* 
    lockfile.c - library module to safely create a lock file
    Copyright (C) 2001  Matthias Andree <matthias.andree@web.de>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "leafnode.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>		/* for kill */

/*
 * check if lockfile exists: if yes (and another instance of fetch is
 * still active) return 1; otherwise generate it. If this fails, also
 * return 1, otherwise return 0. return -1 for other errors.
 *
 * requires: atomic link
 * features: NFS safe (but leafnode does not work distributed)
 * stale detection by PID in lock file (may be fooled)
 */

static char *locktemp;

static void
cleanup(void)
{
    if (locktemp) {
	(void)unlink(locktemp);
	free(locktemp);
	locktemp = 0;
    }
}

static nlink_t
fd_st_nlink(const int fd)
{
/* get nlink count of filedes
   log and return -1 in case of trouble */
    struct stat st;

    if (fstat(fd, &st)) {
	syslog(LOG_ERR, "Cannot fstat %d: %s", fd, strerror(errno));
	return -1;
    }

    return st.st_nlink;
}

static int
lock_is_stale(const char *const name)
{
/* check if the lock file given by "name" is stale
 * i. e. the process no longer exists, erase, if it is stale
 *
 * NOTE: cannot detect if lock files held by other hosts
 * are stale.
 *
 * USAGE: retry to lock only if this 
 *
 * returns: 
 * -1 for error "failure"
 *  0 if lock file is still in use or held by another host "failure"
 *  1 if lock file is stale and has been erased "success"
 */
    char buf[512];
    int fd;
    int len;
    char *pid;
    char *host;
    char *tmp;
    unsigned long npid;

    fd = open(name, O_RDONLY, 0);
    if (fd < 0) {
	if (errno == ENOENT) {
	    /* file has just disappeared, thus it's not stale */
	    return 0;
	} else {
	    syslog(LOG_ERR,
		   "cannot open %s for reading: %s", name, strerror(errno));
	    return -1;
	}
    }

    if ((len = read(fd, buf, sizeof(buf) - 1)) < 0) {
	syslog(LOG_ERR, "read error on %s: %s", name, strerror(errno));
	(void)close(fd);
	return -1;
    }

    if (close(fd) < 0) {
	syslog(LOG_ERR, "read error on %s: %s", name, strerror(errno));
	return -1;
    }

    /* read pid and host */
    buf[len - 1] = '\0';
    pid = host = buf;
    /* we expect a single \n here */
    host += strcspn(host, "\n");
    *(host++) = '\0';

    /* kill trailing \n */
    tmp = host;
    tmp += strcspn(tmp, "\n");
    *tmp = '\0';

    npid = strtoul(pid, 0, 10);
    if (npid == ULONG_MAX && errno == ERANGE) {
	/* overflow error, should not happen, bail out */
	syslog(LOG_ERR, "bogus pid in %s: %s", name, strerror(errno));
	return -1;
    }

    if (strcasecmp(host, fqdn)) {
	syslog(LOG_ERR,
	       "lockfile held by pid %lu on host %s, we're %s",
	       npid, host, fqdn);
	return 0;		/* other host holds the lock */
    }

    /* okay, we can see if there's still a process with that pid active */
    if (kill(npid, 0) && errno == ESRCH) {
	/* no such process, good */
	if (!unlink(name)) {
	    syslog(LOG_NOTICE,
		   "erased stale pid %lu host %s lockfile %s",
		   npid, host, name);
	    return 1;
	} else {
	    syslog(LOG_ERR,
		   "unable to erase stale pid %lu host %s lockfile %s",
		   npid, host, name);
	    return 0;
	}
    }

    /* there is a process active */
    return 0;
}



/* safe mkstemp replacement, ensures the file is only read and writable
 * by owner, some systems create these with 0777 or 0666 permissions. */
int
safe_mkstemp(char *template)
{
    mode_t oldmask;
    int ret;

    oldmask = umask(077);
    ret = mkstemp(template);
    (void)umask(oldmask);

    return ret;
}



int
lockfile_exists(void)
{
/* returns: 
 *  0: if locking succeeded
 *  1: if locking failed because the lock is held by someone else and 
 *     isn't stale
 * -1: for other errors
 */
    char *pid;
    static int registered = 0;
    int fd;
    int have_lock = 0;
    const int block = 0;
    const char *const append = ".XXXXXXXXXX";

    /* kill bogus fqdn */
    if (!strchr(fqdn, '.')
	|| !strncasecmp(fqdn, "localhost", 9)
	|| !strncmp(fqdn, "127.0.0.", 8)
	) {
	syslog(LOG_CRIT,
	       "Internal error: must not try to lock with local host name \"%s\"",
	       fqdn);
	return -1;
    }


    locktemp =
	critmalloc(strlen(lockfile) + strlen(append) + 1, "lockfile_exists");
    pid =
	critmalloc(strlen(fqdn) + sizeof(unsigned long) * 4 +
		   4, "lockfile_exists");

    strcpy(locktemp, lockfile);	/* Flawfinder: ignore */
    strcat(locktemp, append);	/* Flawfinder: ignore */

    if (!registered) {
	if (atexit(cleanup))
	    syslog(LOG_ERR,
		   "cannot register cleanup function, clean up %s.?????????? yourself",
		   lockfile);
	else
	    registered = 1;
    }

    /* make a temporary file */
    fd = safe_mkstemp(locktemp);
    if (fd < 0) {
	syslog(LOG_ERR, "mkstemp(%s) failed: %s", locktemp, strerror(errno));
	free(locktemp);
	locktemp = 0;
	locktemp = 0;
	free(pid);
	return -1;
    }

    /* write our PID and host into it (stale detection) */
    sprintf(pid, "%lu\n%s\n", (unsigned long)getpid(), fqdn);	/* Flawfinder: ignore */
    /* safe, see malloc above */
    if (write(fd, pid, strlen(pid)) < 0 || fsync(fd) < 0) {
	syslog(LOG_ERR, "cannot write to %s: %s", locktemp, strerror(errno));
	if (unlink(locktemp))
	    syslog(LOG_ERR,
		   "Cannot remove lock helper file %s: %s",
		   locktemp, strerror(errno));
	free(locktemp);
	locktemp = 0;
	free(pid);
	return -1;
    }

    /* and try to finally lock */
    while (!have_lock) {
	if (0 == link(locktemp, lockfile)) {
	    /* link succeeded. good. */
	    have_lock = 1;
	    break;
	} else {
	    if (2 == fd_st_nlink(fd)) {
		/* link failed, but st_nlink has increased to 2, good. */
		have_lock = 1;
	    } else {
		int stale;
		struct timeval tv = { 1, 0 };

		/* Could not create link. Check if the lock file is stale. */
		stale = lock_is_stale(lockfile);

		/* if we have a stale file, it's just been removed.
		   retry, don't care for what block says */
		if (stale == 1)
		    continue;

		/* if we have a problem with stale detection, or
		   if we are in non-blocking mode, abort */
		if (stale == -1 || !block)
		    break;

		/* retry after a second, select does not interfere w/ alarm */
		if (select(0, NULL, NULL, NULL, &tv) < 0) {
		    /* must not happen */
		    syslog(LOG_ERR,
			   "lockfile_exists: select failed: %s",
			   strerror(errno));
		    break;
		}
	    }
	}
    }

    if (close(fd) < 0) {
	syslog(LOG_ERR, "cannot write to %s: %s", locktemp, strerror(errno));
	have_lock = 0;
    }

    /* unlink locktemp, but just log if unable to unlink, ignore otherwise */
    if (unlink(locktemp))
	syslog(LOG_ERR,
	       "Cannot remove lock helper file %s: %s",
	       locktemp, strerror(errno));

    /* clean up */
    free(locktemp);
    locktemp = 0;
    free(pid);

    /* mind the return logic */
    return have_lock ? 0 : 1;
}
